/**********************************************************


             Header file for Quick Ray Trace

                     Steve Koren

 **********************************************************/

#include <exec/types.h>
#include <exec/execbase.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef MAXMATH
#include <m68881.h>
#define M020 1
#endif
#ifdef FFP020
#define M020 1
#endif
#ifdef LIB020
#define M020 1
#endif
#define TRUE  1
#define FALSE 0

#define BOOL short


/**********************************************************

                     MACHINE TYPES

 **********************************************************/

#ifndef AMIGA
#define AMIGA          TRUE
#endif
#undef  MAC_II
#undef  ATARI_ST
#undef  UNIX
#undef  IBM_OR_CLONE
#undef  OS_2
#undef  XENIX
#undef  VMS

/* Prototypes for functions defined in bbox.c */
void BboxBbox(struct vector *v1,
              struct vector *v2,
              struct obj_struct *bbox);
void BboxSphere(struct vector *v1,
                struct vector *v2,
                struct obj_struct *sphere);
void BboxParallelogram(struct vector *v1,
                       struct vector *v2,
                       struct obj_struct *para);
void BboxTriangle(struct vector *v1,
                  struct vector *v2,
                  struct obj_struct *obj);
void BboxRing(struct vector *v1,
              struct vector *v2,
              struct obj_struct *ring);
void BboxQuadratic(struct vector *v1,
                   struct vector *v2,
                   struct obj_struct *quad);
/* Prototypes for functions defined in error.c */
void Err(void);
void Warning(int num,
             char *messg);
void Error(int num,
                 int code);
/* Prototypes for functions defined in inout.c */
int LoadWorld(void);
struct obj_struct *Get_Object(void);
void def_colorinfo(struct cinfo_struct *cinfo);
void copy_colorinfo(struct cinfo_struct *c1,
                 struct cinfo_struct *c2);
int Get_Default(void);
int GetOpt(char *str,
           struct cinfo_struct *cinfo);
struct obj_struct *GetSphere(void);
int GetLamp(void);
int GetObserver(void);
struct obj_struct *GetTriangle(void);
int GetSky(void);
struct obj_struct *GetRing(void);
struct obj_struct *GetParallelogram(void);
struct obj_struct *GetQuadratic(void);
int GetFocLength(void);
struct obj_struct *Get_Primitive(char *str);
int GetAttrib(char *str);
/* Prototypes for functions defined in instance.c */
struct obj_struct *Name_Find(struct obj_struct *obj,
                             char *name);
struct obj_struct *Subtree_Copy(struct obj_struct *obj,
                                int fflag);
void Subtree_Offset(struct obj_struct *obj,
                    struct vector *offset,
                    int fflag);
void Subtree_Scale(struct obj_struct *obj,
                   struct vector *mult,
                   int fflag);
struct obj_struct *Get_Instance_Of(void);
/* Prototypes for functions defined in intersect.c */
int LineBbox(struct obj_struct *line,
             struct obj_struct *bbox,
             float *t);
int LineRing(struct obj_struct *line,
             struct obj_struct *ring,
             float *t);
int LineParallelogram(struct obj_struct *line,
                      struct obj_struct *para,
                      float *t);
int LineTriangle(struct obj_struct *line,
                 struct obj_struct *obj,
                 float *t);
int LineSphere(struct obj_struct *line,
               struct obj_struct *sph,
               float *t);
int LineQuadratic(struct obj_struct *line,
                  struct obj_struct *quad,
                  float *t);
/* Prototypes for functions defined in lexer.c */
char towhite(int c);
void rmspace(void);
void Comment_Killer(void);
void GetToken(char *s);
float InRange(float cnum);
float IsPos(float val);
float Get_Next_Num(void);
char *Get_Next_Name(void);
short Get_Color_Val(void);
void GetLeftParen(void);
int GetRightParen(void);
void GetVector(struct vector *vector);
void GetSVector(struct svector *svector);
/* Prototypes for functions defined in mth.c */
int PsRand(void);
void VectEqZero(struct vector *v);
void VectAddMult(struct vector *v1,
                 float k2,
                 struct vector *v2,
                 float k3,
                 struct vector *v3);
void VectNegate(struct vector *v1,
                struct vector *v2);
void VectorAdd(struct vector *v1,
               struct vector *v2,
               struct vector *v3);
void VectorMult(struct vector *v1,
                struct vector *v2,
                struct vector *v3);
void VecSubtract(struct vector *v1,
                 struct vector *v2,
                 struct vector *v3);
void VectScale(struct vector *v1,
               float k,
               struct vector *v2);
void SVectEQ(struct svector *v1,
             struct svector *v2);
void VectEQ(struct vector *v1,
            struct vector *v2);
void Normalize(struct vector *v);
void CrossProd(struct vector *v1,
               struct vector *v2,
               struct vector *v3);
void FindPos(struct vector *pos,
             struct obj_struct *line2,
             float t);
void Reflect(struct vector *out,
             struct vector *in,
             struct vector *norm);
float atan2w(float x,
             float y);
void Rot12(struct vector *invect,
           struct vector *outvect,
           register float cos1,
           register float sin1,
           register float cos2,
           register float sin2);
void Rot21(struct vector *invect,
           struct vector *outvect,
           float cos1,
           float sin1,
           float cos2,
           float sin2);
/* Prototypes for functions defined in norm.c */
void SphereNorm(struct vector *norm,
                struct obj_struct *object,
                struct vector *position);
void PlaneNorm(struct vector *norm,
               struct obj_struct *object,
               struct vector *position);
void QuadraticNorm(struct vector *norm,
                   struct obj_struct *object,
                   struct vector *position);
/* Prototypes for functions defined in offset.c */
void Standard_Offset(struct obj_struct *obj,
                     struct vector *offset);
void Offset_Bbox(struct obj_struct *obj,
                 struct vector *offset);
/* Prototypes for functions defined in pattern.c */
int Find_Color(struct obj_struct *obj,
               struct patt *pattern,
               struct vector *loc,
               struct cinfo_struct *cinfo,
               float xmult,
               float ymult);
int Rect_Hit(float pos1,
             float pos2,
             struct patt *patt);
int Circle_Hit(float pos1,
               float pos2,
               struct patt *patt);
int line_intersect(float px,
                   float py,
                   float x1,
                   float y1,
                   float x2,
                   float y2);
int Poly_Hit(float pos1,
             float pos2,
             struct patt *patt);
/* Prototypes for functions defined in patternio.c */
struct patt *Attach_Pattern(void);
struct patt *Get_Circle_Pattern(void);
struct patt *Get_Rect_Pattern(void);
struct patt *Get_Poly_Pattern(void);
struct patt *Get_SubPattern(char *str);
int GetPattern(void);
/* Prototypes for functions defined in precomp.c */
void PreCompSphere(struct obj_struct *obj);
void PreCompPlane(struct obj_struct *obj);
void PreCompQuadratic(struct obj_struct *obj);
void PreCompNull(struct obj_struct *obj);
/* Prototypes for functions defined in qrt.c */
void Setup_Observer(void);
void init_world(void);
int main(int argc,
          char **argv);
void Parse_CL_Args(int argc,
                 char **argv);
/* Prototypes for functions defined in ray.c */
void DiffColor(struct svector *color,
               struct cinfo_struct *cinfo,
               struct vector *norm,
               struct vector *loc,
               struct obj_struct *oline);
void AmbColor(struct svector *color,
              struct cinfo_struct *cinfo,
              struct vector *norm,
              struct vector *loc);
void TransColor(struct svector *color,
                struct cinfo_struct *cinfo,
                struct vector *norm,
                struct vector *loc,
                struct obj_struct *line,
                float inmult);
void ReflectColor(struct svector *color,
                  struct cinfo_struct *cinfo,
                  struct vector *norm,
                  struct vector *loc,
                  struct obj_struct *line,
                  float inmult);
void SkyColor(struct obj_struct *line,
              struct svector *color);
void Dither(struct svector *color,
                 struct cinfo_struct *cinfo);
struct obj_struct *Ray_Hit(struct obj_struct *CurrObj,
                           struct obj_struct *line,
                           float *MinT,
                           short sflag,
                           short fflag,
                           struct vector *atten);
int Ray_Trace(struct obj_struct *line,
              struct svector *color,
              float multiplier);
void PixelLine(int x,
               int y,
               struct obj_struct *line);
void Dump_Line(int lineno,
               short *r,
               short *g,
               short *b);
void Screen_Trace(void);
/* Prototypes for functions defined in relpos.c */
void Find_Plane_Coords(struct vector *v1,
                       struct vector *v2,
                       struct vector *delta,
                       float *pos1,
                       float *pos2,
                       float t1,
                       float t2,
                       float t3);
void Plane_Pos(struct obj_struct *obj,
               struct vector *loc,
               float *pos1,
               float *pos2,
               int normalize);
void Sphere_Pos(struct obj_struct *obj,
                struct vector *loc,
                float *pos1,
                float *pos2);
void Quadratic_Pos(struct obj_struct *obj,
                   struct vector *loc,
                   float *pos1,
                   float *pos2);
/* Prototypes for functions defined in resize.c */
void Resize_Sphere(struct obj_struct *obj,
                   struct vector *mult);
void Resize_Plane(struct obj_struct *obj,
                  struct vector *mult);
void Resize_Quadratic(struct obj_struct *obj,
                      struct vector *mult);
void Resize_Bbox(struct obj_struct *obj,
                 struct vector *mult);
/* Prototypes for functions defined in stack.c */
void Do_Precomp(struct obj_struct *node);
void Do_Precomp_Tree(struct obj_struct *node);
void Make_Bbox(struct obj_struct *node);
struct patt *find_pat(char *name);
struct patt *new_pat(void);
struct obj_struct *new_obj(short type,
                           struct vector *loc,
                           struct vector *v1,
                           struct vector *v2,
                           struct vector *v3,
                           struct cinfo_struct *cinfo,
                           struct patt *pattern,
                           struct patt *remove,
                           char *name,
                           struct vector *upper,
                           struct vector *lower,
                           float cterm,
                           float xmult,
                           float ymult);
struct obj_struct *new_line(void);
void add_lamp(struct obj_struct *object);
void World_Stats(void);
void Open_File(void);
void Close_File(void);

/**********************************************************

                   MISC MATH CONSTANTS

 **********************************************************/

#ifndef PI
#define PI  3.141592
#endif
#define PI2 1.570796


/**********************************************************

                     OBJECT  NUMBERS

 **********************************************************/

#define LINE               1
#define SPHERE             2
#define PARALLELOGRAM      3
#define TRIANGLE           4
#define LAMP               5
#define OBSERVER           6
#define GROUND             7
#define SKY                8
#define BBOX               9
#define RING               10
#define QUADRATIC          11


/**********************************************************

                   PROGRAM CONSTANTS

 **********************************************************/

#define SMALL    1.0e-3      /* a sorta small number    */
#define BIG      1.0e15      /* a sorta big number      */
#define CNUM     100         /* this many shades/color  */
#define SLEN     64          /* max string length       */
#define MAX_IX   4           /* maximum x interpolation */
#define MAX_IY   4           /* maximum y interpolation */
#define MAX_XRES 720         /* maximum X resolution    */
#define XYZZY    42          /* this is here for no     */
                             /* reason whatsoever       */


/**********************************************************

                      VECTOR STRUCTURES

 **********************************************************/

typedef struct vector {        /* a vector in 3 space   */
  float x,y,z;
} VECTOR, *VECT_PTR;

typedef struct svector {       /* an r,g,b color vector */
  short r,g,b;
} SVECTOR, *SVECT_PTR;

typedef struct cinfo_struct {  /* color information     */

  SVECTOR       amb,           /* ambient lighting      */
                diff,          /* diffuse lighting      */
                mirror,        /* % light reflected     */
                trans;         /* % light transmitted   */

  VECTOR        density;       /* density */

  float         sreflect,      /* specular refl coefficient */
                index;         /* index if refraction */

  short         fuzz,          /* currently unused */
                reflect,       /* percent specularly reflected */
                dither;        /* color dithering. 3..6 look ok */

} CINFO, *CINFO_PTR;


/**********************************************************

                PRECOMPUTED INFO FOR OBJECTS

 These fields can be used by object routines however
 they wish.  They just make object/line intersections
 faster.

 **********************************************************/

typedef struct _pre {

  float    sin1, cos1,         /* sin and cos */
           sin2, cos2,
           n1,                 /* misc number */
           len1, len2;         /* lengths of vectors */

  VECTOR   vect1,              /* misc vector */
           norm;               /* norm for planar objs */

} PRECOMP, PRECOMP_PTR;


/**********************************************************

                      PATTERN STRUCTURE

 **********************************************************/

typedef struct patt {
  short    type;               /* type of pattern */
  float    xsize,              /* pattern size */
           ysize,
           startx,
           starty,             /* x,y positions */
           endx,
           endy,
           radius;             /* rad for circles */

  CINFO    cinfo;              /* color information */

  char     *name;              /* pattern name */

  struct patt *child, *sibling, *link;

} PATTERN, *PATTERN_PTR;


/**********************************************************

                    OBJECT STRUCTURE

 **********************************************************/

typedef struct obj_struct {

  short         type,         /* object type */
                flag;         /* misc boolean flag */

  char          *name;        /* object name */

  VECTOR        loc,          /* object location */
                vect1,        /* three vectors */
                vect2,
                vect3,
                lower,        /* lower and upper bounds */
                upper;

  float         cterm,        /* for quadratic surfaces only */
                xmult,        /* x and y multipliers for patterns */
                ymult;

  CINFO         cinfo;        /* color information */

  PRECOMP       precomp;      /* precomputed information */

  struct obj_struct *nextobj, /* next obj in list */
                    *child;   /* child for bounding boxes only */

  PATTERN_PTR   pattern,      /* pointer to pattern structure */
                remove;       /* remove section of object */

} OBJ_STRUCT, *OBJ_PTR;


/**********************************************************

                   Plane Bbox structure

 **********************************************************/

typedef struct _PlaneBox {
  int      min_x, min_y,
           max_x, max_y;

  OBJ_PTR  object;

  struct _PlaneBox *next;
} PLANE_BBOX, *PLANE_BBOX_PTR;


/**********************************************************

                       WORLD STRUCTURE

 **********************************************************/

typedef struct world {
  OBJ_PTR   stack,            /* here are the objects */
            observer,         /* the observer */
            sky,              /* sky */
            lamps,            /* a lamp list */
            instances;        /* instance list */

  int       objcount,         /* # objects and lamps */
            lampcount;

  long      ray_intersects,   /* statistics */
            primary_traced,
            to_lamp,
            refl_trans,
            bbox_intersects,
            intersect_tests,
            pixels_hit,
            pattern_matches;

  VECTOR    obsright,         /* obs up dir */
            obsup;

  SVECTOR   skycolor_horiz,   /* skycolors */
            skycolor_zenith;

  PATTERN_PTR patlist;        /* the pattern stack */

  float     flength,          /* focal length */
            x_divisor,        /* used to find ray direction */
            y_divisor,
            globindex;        /* global index of refraction */

  char      *outfile;         /* output file name */
  FILE      *filept;          /* output file pointer */
} WORLD;


/**********************************************************

                  FUNCTIONS FOR OBJECT TYPES

 **********************************************************/

typedef struct obj_data {
  int (*ColTest)();           /* collision test function ptr */
  int (*FindNorm)();          /* normal finding function ptr */
  int (*FindBbox)();          /* object bound function ptr   */
  int (*RelPos)();            /* object relative position    */
  int (*PreComp)();           /* info pre-computing routine  */
  int (*Offset)();            /* offset object by dx, dy, dz */
  int (*Resize)();            /* resize object by a multiple */
} OBJ_DATA;


/**********************************************************

                      MATH DEFINES

 **********************************************************/

#define sqr(x) ((x)*(x))
#define DotProd(v1,v2) (v1.x*v2.x+v1.y*v2.y+v1.z*v2.z)
#define MIN(x,y) ((x)<(y) ? (x) : (y))
#define MAX(x,y) ((x)>(y) ? (x) : (y))
#define IABS(x)  ((x)>0   ? (x) : (-(x)))


/**********************************************************

                      Default structure

 **********************************************************/

typedef struct def_struct {

  CINFO cinfo;             /* default colorinfo */

  short shadow,            /* shadows ? */
        vlamp,             /* lamps visible (not yet implimented) */
        int_x,             /* interpolate (def=1) */
        int_y,

        x_res,             /* X resolution of image */
        y_res,             /* Y resolution of image */
        x_center,          /* X center of image */
        y_center;          /* Y center of image */

  float threshold,         /* cutoff pt for min refl, refl rays */
        aspect;            /* aspect ratio for image            */


  short ithreshold;        /* integer version of above          */

} DEF, *DEF_PTR;


/**********************************************************

                     EXTERNAL DECLARATIONS

 **********************************************************/

extern DEF def;

extern OBJ_DATA ObjData[];

extern WORLD THEWORLD;


/**********************************************************

                       ERROR CODES

 **********************************************************/

#define ILLEGAL_PARAMETER 1
#define TOO_FEW_PARMS     2
#define ILLEGAL_OBJECT    3
#define MALLOC_FAILURE    4
#define SYNTAX_ERROR      5
#define INTERNAL_ERROR    6
#define FILE_ERROR        7
#define PATTERN_NOT_FOUND 8
#define PATTERN_EXISTS    9
#define NO_OBSERVER       10
#define UNDEFINED_PARAM   11
#define NON_HOMOGENIOUS   12
#define ZERO_INDEX        13
#define COLOR_VALUE_ERR   14
#define LESS_THAN_ZERO    15
#define ZERO_MULTIPLIER   16
#define UNDEFINED_NAME    17
#define LPAREN_EXPECTED   18
#define RPAREN_EXPECTED   19
#define ILLEGAL_VECTOR    20
#define ILLEGAL_SVECTOR   21
#define ILLEGAL_OPTION    22


/**********************************************************

                      WARNING CODES

 **********************************************************/

#define OBSOLETE_OPTION   1


/**********************************************************

 Define this flag for more robust code (HIGHLY recommended)

 **********************************************************/

#define ROBUST TRUE


