/************************************************************************
 *									*
 *		    Copyright (c) 1987, David B. Wecker			*
 *			    All Rights Reserved				*
 *									*
 * This file is part of DBW_Render					*
 *									*
 * DBW_Render is distributed in the hope that it will be useful, but	*
 * WITHOUT ANY WARRANTY. No author or distributor accepts		*
 * responsibility to anyone for the consequences of using it or for	*
 * whether it serves any particular purpose or works at all, unless	*
 * he says so in writing. Refer to the DBW_Render General Public	*
 * License for full details.						*
 *									*
 * Everyone is granted permission to copy, modify and redistribute	*
 * DBW_Render, but only under the conditions described in the		*
 * DBW_Render General Public License. A copy of this license is		*
 * supposed to have been given to you along with DBW_Render so you	*
 * can know your rights and responsibilities. It should be in a file	*
 * named COPYING. Among other things, the copyright notice and this	*
 * notice must be preserved on all copies.				*
 ************************************************************************
 *									*
 * Authors:								*
 *	DBW - David B. Wecker						*
 *									*
 * Versions:								*
 *	V1.0 870125 DBW	- First released version			*
 *									*
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <exec/types.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/exec.h>

#define VERSION "RAY2 v1.0 870125 (C) 1987 D. Wecker - all rights reserved\n"
#define BPP	    4			/* Bits per pixel */
#define MAXROW	    400			/* Rows run 0-399 */
#define MAXCOL	    320			/* Cols run 0-319 */
#define MAXBYTE	    40			/* MAXCOL / 8     */
#define DEPTH	    6
#define MAXGRAY	    (1 << BPP)
#define PPW	    (sizeof( int ) * 8 / BPP)
#define WPSL	    (MAXCOL / PPW)
#define	ABS(x)	    abs(x)

#define BMHDsize    20L			/* chunk sizes */
#define CMAPsize    96L
#define CAMGsize    4L
#define BODYsize    ((long)((16000L >> (1-lace)) * (long)(depth)))
#define FORMsize    (BODYsize+CAMGsize+CMAPsize+BMHDsize+36L)

/* Prototypes for functions defined in ray2.c */
void error(int lev,
           char *msg);
unsigned char *PutDump(unsigned char *dest,
                       int nn);
unsigned char *PutRun(unsigned char *dest,
                      int nn,
                      int cc);
int PackRow(unsigned char **pSource,
            unsigned char **pDest,
            int RowSize);
int read_scanline(int pass,
                  int row);
int main(int argc,
         char **argv);
int getcolor(int ppix,
             int *crgb,
             int prgb);
int distance(int argb,
             int brgb);
struct	RastPort	*rp;
struct	ViewPort	*vp;
struct	Window		*w;
struct	Screen		*screen;
struct	IntuiMessage	*message;
struct	NewScreen	ns = {
    0L,0L,320L,400L,6L,
    0,1,HAM|LACE,
    CUSTOMSCREEN,NULL,
    (UBYTE *)"DBW_RENDER2",
    NULL,NULL };
struct	NewWindow	nw = {
    0L,0L,320L,400L,0L,1L,
    MOUSEBUTTONS|CLOSEWINDOW,
    ACTIVATE|WINDOWCLOSE|BORDERLESS,
    NULL,NULL,
    (UBYTE *)"DBW_RENDER2",
    NULL,NULL,
    0L,0L,320L,400L,CUSTOMSCREEN };

typedef int	    scanlinetype[3][WPSL];
scanlinetype	    line;

struct {
    int	    curr;
    FILE    *fp;
    }		    ifil[10];
static int	    numfiles = 0;
FILE		    *ofil;
static int	    comp = 1;
static int	    depth  = 6;
static int	    colors = 16;
static int	    ham = 1;
static int	    lace = 1;
static int	    colorstats[4096][2];
static int	    image[MAXCOL];
char		    fname[40],str[40];
long		    lng,pos1,pos2;
int		    wrd;
unsigned char	    byt;
unsigned long	    class;
unsigned short	    code;
unsigned char	    *dest,destbuf[MAXCOL+10],*planes[DEPTH];

void error(lev,msg)
int   lev;
char *msg;
    {
    switch (lev) {
	case 6:

	CloseWindow(w);

	case 5:
	CloseScreen(screen);

	case 4:
	CloseLibrary((struct Library *)IntuitionBase);

	case 3:
	CloseLibrary(GfxBase);

	case 2:
	if (ofil != NULL) fclose(ofil);

	case 1:
	while (numfiles > 0) fclose(ifil[--numfiles].fp);

	case 0:
	if (msg) {
	    printf("\nUsage: ray2 [-c #] [-h #] [-l #] [-o ofil] ifil\n");
	    printf("ERROR: %s\n",msg);
	    exit(0);
	    }
	}
    exit(0);
    }

/************************ run length encoding from RKM *****************/
#define DUMP		0
#define RUN		1
#define MinRun		3
#define MaxRun		128
#define	MaxDat		128
#define GetByte()	(*source++)
#define PutByte(c)	{ *dest++ = (c); ++PutSize; }
#define OutDump(nn)	dest = PutDump(dest,nn);
#define OutRun(nn,cc)	dest = PutRun(dest,nn,cc);

int	PutSize;
char	buf[256];

unsigned char *PutDump(dest,nn)
unsigned char	*dest;
int		nn;
    {
    int i;

    PutByte(nn-1);
    for (i = 0; i < nn; i++) PutByte(buf[i]);
    return(dest);
    }

unsigned char *PutRun(dest,nn,cc)
unsigned char	*dest;
int		nn,cc;
    {
    PutByte(-(nn-1));
    PutByte(cc);
    return(dest);
    }

int PackRow(pSource,pDest,RowSize)
unsigned char	**pSource,**pDest;
int		RowSize;
    {
    unsigned char   *source,*dest;
    char	    c,
		    lastc = '\000';
    int		    mode = DUMP,
		    nbuf = 0,	    /* number of chars in buf */
		    rstart = 0;	    /* buf index current run starts */

    source = *pSource;
    dest   = *pDest;

    PutSize = 0;
    buf[0]  = lastc = c = GetByte();
    nbuf    = 1;
    RowSize--;

    for (; RowSize; --RowSize) {
	buf[nbuf++] = c = GetByte();

	switch (mode) {
	    case DUMP:
	    if (nbuf > MaxDat) {
		OutDump(nbuf-1);
		buf[0]	= c;
		nbuf	= 1;
		rstart	= 0;
		break;
		}
	    if (c == lastc) {
		if (nbuf-rstart >= MinRun) {
		    if (rstart > 0) OutDump(rstart);
		    mode = RUN;
		    }
		else if (rstart == 0) mode = RUN;
		}
	    else rstart = nbuf - 1;
	    break;

	    case RUN:
	    if ((c != lastc) || (nbuf-rstart > MaxRun)) {
		OutRun((nbuf-1)-rstart,lastc);
		buf[0]	= c;
		nbuf	= 1;
		rstart	= 0;
		mode	= DUMP;
		}
	    break;
	    }
	lastc = c;
	}
    switch (mode) {
	case DUMP: OutDump(nbuf); break;
	case RUN:  OutRun(nbuf-rstart,lastc); break;
	}

    *pSource = source;
    *pDest   = dest;

    return(PutSize);
    }

/******************* end of RKM encoding routines **********************/

read_scanline(pass,row)
int	pass,row;
    {
    int i,j,x,val,good;

    if ((row % 60) == 0) printf("\nRow: %4d ",row);
    printf(".");
    fflush(stdout);

    good = 0;
    for (i = 0; i < numfiles && good == 0; i++) {
	while (ifil[i].curr < row) {
	    if (fread((char *)&ifil[i].curr,1,sizeof(int),ifil[i].fp)!=sizeof(int)) {
		ifil[i].curr = 9999; 
		continue;
		}
	    if (fread((char *)line,1,sizeof(scanlinetype),ifil[i].fp)!=sizeof(scanlinetype)) {
		ifil[i].curr = 9999;
		continue;
		}
	    if (ifil[i].curr == row) {
		good = 1;
		break;
		}
	    }
	}

    /* couldn't find this scan line in any file */
    if (good == 0) return(0);

    x = 0;
    for (i = 0; i < WPSL; i++) {
	for (j = 0; j < PPW; j++) {
	    val =
		(((line[0][i] >> (BPP * j)) & (MAXGRAY-1)) << (BPP*2)) +
		(((line[1][i] >> (BPP * j)) & (MAXGRAY-1)) <<  BPP)    +
		 ((line[2][i] >> (BPP * j)) & (MAXGRAY-1));

	    val %= 4096;
	    if (pass == 1) {
		if (colorstats[val][1] < 32767) colorstats[val][1]++;
		}
	    else image[x++] = val;
	    }
	}
    return(1);
    }

int main(argc,argv)
int	argc;
char	**argv;
    {
    int i,j,k,gap,t0,t1,prgb,crgb,cpix,ppix,maxdis,used;

    printf(VERSION);

    fname[0] = '\000';
    for (i=1; i<argc; i++) {
	if	(strcmp(argv[i],"-o") == 0) {
	    if (++i < argc) strcpy(fname,argv[i]);
	    else	    error(0,"Bad argument to output file command");
	    }
	else if (strcmp(argv[i],"-c") == 0) {
	    if (++i >= argc) error(0,"Need parameter to compression");
	    if (sscanf(argv[i],"%d",&comp) != 1 || (comp != 0 && comp != 1))
		error(0,"compression switch must be followed by a 0 or 1");
	    }
	else if (strcmp(argv[i],"-h") == 0) {
	    if (++i >= argc) error(0,"Need parameter to ham switch");
	    if (sscanf(argv[i],"%d",&ham) != 1  || (ham != 0 && ham != 1))
		error(0,"ham switch must be 0 or 1");
	    if (ham == 0) { depth = 5; colors = 32;}
	    else	  { depth = 6; colors = 16; lace = 1;}
	    }
	else if (strcmp(argv[i],"-l") == 0) {
	    if (++i >= argc) error(0,"Need parameter to lace switch");
	    if (sscanf(argv[i],"%d",&lace) != 1  || (lace != 0 && lace != 1))
		error(0,"lace switch must be 0 or 1");
	    if (lace == 0) { depth = 5; colors = 32; ham = 0; }
	    }
	else break;
	}

    if (i >= argc)
	error(0,
	"Must supply the root name of at least one .TMP file from RENDER");
    while (i < argc) {
	if (numfiles > 9) error(1,"too many input files supplied");
	sprintf(str,"%s.tmp" ,argv[i]);
	if (fname[0] == '\000') sprintf(fname,"%s.ilbm",argv[i]);
	ifil[numfiles].fp = fopen(str,"r");
	if (ifil[numfiles].fp == NULL)
	    error(1,"opening an input file, supply names with no extensions");
	ifil[numfiles].curr = -9999;
	numfiles++;
	i++;
	}

    ofil = fopen(fname,"w");
    if (ofil == NULL) error(2,"creating output file");

    printf("Building a color histogram:\n");
    for (i = 0; i < 4096; i++) {
	colorstats[i][0] = i;
	colorstats[i][1] = 0;
	}
    k = 0;
    for (i = 0; i < 400; i++) read_scanline(1,i);
    printf("\n");
    for (i = 0; i < numfiles; i++) {
	fseek(ifil[i].fp,0L,0);
	ifil[i].curr = -9999;
	}

    for (gap = 2048; gap > 0; gap /=2) {
	for (i = gap; i < 4096; i++) {
	    for (j = i-gap; j >= 0 && 
		colorstats[j][1] < colorstats[j+gap][1]; j -= gap) {
		t0		    = colorstats[j][0];
		t1		    = colorstats[j][1];
		colorstats[j][0]    = colorstats[j+gap][0];
		colorstats[j][1]    = colorstats[j+gap][1];
		colorstats[j+gap][0]= t0;
		colorstats[j+gap][1]= t1;
		}
	    }
	}

    for (used = 0; used < 4096 && colorstats[used][1] > 0; used++) ;
    printf("\nUsed %d colors out of a possible 4096\n",used);

    /* if we're not in HAM mode, compress the color table */
    if (ham == 0 && used > colors) {
	for (maxdis = 2; maxdis < 1024; maxdis *= 2) {
	    for (i=colors/2; i < colors; i++) {
		for (j = 0; j < i; j++) {
		    if (distance(colorstats[i][0],colorstats[j][0])<maxdis) {
			for (j=i+1; j<used; j++) {
			    colorstats[j-1][0] = colorstats[j][0];
			    colorstats[j-1][1] = colorstats[j][1];
			    }
			--used;
			--i;
			break;
			}
		    }
		if (used <= colors) break;
		}
	    if (used <= colors) break;
	    }
	printf("Colors: ");
	for (i = 0; i < colors; i++) {
	    printf("%3x ",colorstats[i][0]);
	    if ((i % 15) == 14) printf("\n\t");
	    }
	printf("\n");
	}

    /* set up to display the image */
    GfxBase	= (struct GfxBase *)OpenLibrary("graphics.library",0L);
    if (GfxBase == NULL) error(2,"No graphics library");
    IntuitionBase = 
	(struct IntuitionBase *)OpenLibrary("intuition.library",0L);
    if (IntuitionBase == NULL) error(3,"No intuition library");

    if (ham == 0) {
	ns.Depth	= (long)depth;
	if (lace) ns.ViewModes = LACE;
	else {
	    ns.ViewModes = 0;
	    ns.Height	 = 200L;
	    nw.Height	 = 200L;
	    nw.MaxHeight = 200L;
	    }
	}
    screen	= (struct Screen *)OpenScreen(&ns);
    if (screen == NULL) error(4,"Can't open screen");
    nw.Screen	= screen;
    w		= (struct Window *)OpenWindow(&nw);
    if (w == NULL) error(5,"Can't open window");
    vp		= &screen->ViewPort;
    rp		= w->RPort;

    /* now set the color map */
    for (i = 0; i < colors; i++) {
	SetRGB4(vp, (long)i, (long)((colorstats[i][0] >> 8) & 0xF),
			     (long)((colorstats[i][0] >> 4) & 0xF),
			     (long)((colorstats[i][0]     ) & 0xF));
	}

    /* now figure out the pixels to display */
    cpix    = 0;
    crgb    = colorstats[0][1];
    SetDrMd(rp,JAM1);

    for (i = 0; i < MAXROW; i += (2-lace)) {

	/* check if there was an attempt to abort */
	while ((message=(struct IntuiMessage *)GetMsg(w->UserPort))!=NULL) {
	    class   = message->Class;
	    code    = message->Code;
	    ReplyMsg((struct Message *)message);

	    if (class == CLOSEWINDOW) error(6,"ABORTED");
	    }

	if (read_scanline(2,i)) {
	    for (j = 0 ; j < MAXCOL; j++) {
		prgb = crgb;
		crgb = image[j];
		ppix = cpix;

		/* start of scan line is ALWAYS an absolute color */
		/* so is a window that is NOT in HAM mode */

		if (ham == 0 || j == 0)	cpix = getcolor(ppix,&crgb,-1);
		else			cpix = getcolor(ppix,&crgb,prgb);

		/* display the computed pixel */
		SetAPen(rp,(long)cpix);
		if (lace) WritePixel(rp,(long)j,(long)i);
		else	  WritePixel(rp,(long)j,(long)(i/2));
		}
	    }
	}
    printf("\n");

    /* here is where we write the ILBM file out */
    sprintf(str,"FORM");    fwrite( str,1,4,ofil);
    pos1 = ftell(ofil);
    lng = FORMsize;	    fwrite((char *)&lng,4,1,ofil);
    sprintf(str,"ILBM");    fwrite( str,1,4,ofil);

    sprintf(str,"BMHD");    fwrite( str,1,4,ofil);
    lng = BMHDsize;	    fwrite((char *)&lng,4,1,ofil);
    wrd = MAXCOL;	    fwrite((char *)&wrd,2,1,ofil);   /* width */
    if (lace) wrd = MAXROW;
    else      wrd = MAXROW/2;
			    fwrite((char *)&wrd,2,1,ofil);   /* height */
    wrd = 0;		    fwrite((char *)&wrd,2,1,ofil);   /* top */
    wrd = 0;		    fwrite((char *)&wrd,2,1,ofil);   /* left */
    byt = depth;	    fwrite(&byt,1,1,ofil);   /* Depth */
    byt = 0;		    fwrite(&byt,1,1,ofil);   /* mask */
    byt = comp;		    fwrite(&byt,1,1,ofil);   /* compress */
    byt = 0;		    fwrite(&byt,1,1,ofil);   /* pad */
    wrd = 0;		    fwrite((char *)&wrd,2,1,ofil);   /* transparency */
    byt = 10;		    fwrite(&byt,1,1,ofil);   /* aspect x */
    byt = 11;		    fwrite(&byt,1,1,ofil);   /* aspect y */
    wrd = MAXCOL;	    fwrite((char *)&wrd,2,1,ofil);   /* page width */
    if (lace) wrd = MAXROW;
    else      wrd = MAXROW/2;
			    fwrite((char *)&wrd,2,1,ofil);   /* page height */

    sprintf(str,"CAMG");    fwrite( str,1,4,ofil);
    lng = CAMGsize;	    fwrite((char *)&lng,4,1,ofil);
    if (ham)	    lng = HAM|LACE;
    else if (lace)  lng = LACE;
    else	    lng = 0L;
			    fwrite((char *)&lng,4,1,ofil);

    sprintf(str,"CMAP");    fwrite( str,1,4,ofil);
    lng = CMAPsize;	    fwrite((char *)&lng,4,1,ofil);
    for (i = 0; i < 32; i++) {
	str[0] = (colorstats[i][0] >> 4) & 0xF0;
	str[1] = (colorstats[i][0]     ) & 0xF0;
	str[2] = (colorstats[i][0] << 4) & 0xF0;
	fwrite( str,1,3,ofil);
	}

    sprintf(str,"BODY");    fwrite( str,1,4,ofil);
    pos2 = ftell(ofil);
    lng = BODYsize;	    fwrite((char *)&lng,4,1,ofil);

    lng = 0L;
    for (i = 0; i < depth; i++)
	planes[i] = (unsigned char *)rp->BitMap->Planes[i];

    for (i = 0; i < MAXROW; i += (2-lace)) {
	/* check if there was an attempt to abort */
	while ((message=(struct IntuiMessage *)GetMsg(w->UserPort))!=NULL) {
	    class   = message->Class;
	    code    = message->Code;
	    ReplyMsg((struct Message *)message);

	    if (class == CLOSEWINDOW) error(6,"ABORTED");
	    }

	for (j = 0; j < depth; j++) {
	    if (comp) {
		dest	    = destbuf;
		wrd	    = PackRow(&planes[j],&dest,MAXBYTE);
		lng	   += (long)wrd;
		fwrite(destbuf,1,wrd,ofil);
		}
	    else {
		fwrite(planes[j],1,MAXBYTE,ofil);
		planes[j] += MAXBYTE;
		}
	    }
	if (lace)   cpix  = (int)ReadPixel(rp,0L,(long)i);
	else	    cpix  = (int)ReadPixel(rp,0L,(long)(i/2));
	cpix ^= colors-1;
	SetAPen(rp,(long)cpix);
	if (lace) WritePixel(rp,0L,(long)i);
	else	  WritePixel(rp,0L,(long)(i/2));
	}
    if (comp) {
	fseek(ofil,(long)pos2,0); fwrite((char *)&lng,4,1,ofil);
	lng -= BODYsize;
	lng += FORMsize;
	fseek(ofil,(long)pos1,0); fwrite((char *)&lng,4,1,ofil);
	}
    fclose(ofil);

    /* flash that we're done */
    DisplayBeep(0L);
    DisplayBeep(0L);

    /* wait here until we get a close window message */
    while (1) {
	WaitPort(w->UserPort);
	while ((message=(struct IntuiMessage *)GetMsg(w->UserPort))!=NULL) {
	    class   = message->Class;
	    code    = message->Code;
	    ReplyMsg((struct Message *)message);

	    if (class == CLOSEWINDOW) error(6,NULL);
	    }
	}
    error(6,NULL);
    }

/* get the next encoding for a pixel */
getcolor(ppix,crgb,prgb)
int ppix,*crgb,prgb;
    {
    int i,j,val,cr,cg,cb,pr,pg,pb,nr,ng,nb,best,dist,nrgb;

    /* if same color, then do a NOOP (same as previous pixel) */
    if (*crgb == prgb) return(ppix);

    /* set up for comparisons */
    cb	= *crgb & (MAXGRAY-1);
    cg	= (*crgb >> BPP) & (MAXGRAY-1);
    cr	= (*crgb >> (BPP*2)) & (MAXGRAY-1);

    pb	= prgb & (MAXGRAY-1);
    pg	= (prgb >> BPP) & (MAXGRAY-1);
    pr	= (prgb >> (BPP*2)) & (MAXGRAY-1);

    /* see if only one plane changed */
    if (prgb != -1) {
	if (pr == cr && pg == cg) return (cb + 0x10);
	if (pr == cr && pb == cb) return (cg + 0x30);
	if (pg == cg && pb == cb) return (cr + 0x20);
	}

    /* look for an exact match in the color table (or minimal distance) */
    for (i=0; i < colors; i++) {
	if (colorstats[i][0] == *crgb) return (i);
	if (i == 0) {
	    best = 0;
	    dist = distance(colorstats[i][0],*crgb);
	    }
	else if ((j=distance(colorstats[i][0],*crgb)) < dist) {
	    best = i;
	    dist = j;
	    }
	}

    /* do a forced absolute */
    if (prgb == -1) {
	*crgb = colorstats[best][0];
	return(best);
	}

    /* find which color is off the most from previous */
    i	= 0;
    val = ABS(cr-pr);
    if (ABS(cg-pg) > val) { i = 1; val = ABS(cg-pg); }
    if (ABS(cb-pb) > val) { i = 2; val = ABS(cb-pb); }

    nr = pr;
    ng = pg;
    nb = pb;
    switch (i) {
	case 0:	nr = cr; val = nr + 0x20; break;
	case 1: ng = cg; val = ng + 0x30; break;
	case 2:	nb = cb; val = nb + 0x10; break;
	}
    nrgb = (nr << (2*BPP)) + (ng << BPP) + nb;

    /* now pick the best */
    if (distance(nrgb,*crgb) < dist) {

	/* do a best relative */
	*crgb = nrgb;
	return(val);
	}

    /* do a best absolute */
    *crgb = colorstats[best][0];
    return(best);
    }

distance(argb,brgb) {
    int b,g,r;

    /* set up for comparisons */
    b	= argb & (MAXGRAY-1);
    g	= (argb >> BPP) & (MAXGRAY-1);
    r	= (argb >> (BPP*2)) & (MAXGRAY-1);

    b	-= brgb & (MAXGRAY-1);
    g	-= (brgb >> BPP) & (MAXGRAY-1);
    r	-= (brgb >> (BPP*2)) & (MAXGRAY-1);

    return(r*r + g*g + b*b);
    }
